"""GitHub webhooks routing proxy."""

# pylint: disable=unused-import
from ...routing.default_router import (  # noqa: F401
    dispatch_event,
    process_event,
    process_event_actions,
    WEBHOOK_EVENTS_ROUTER,
)



"""GitHub webhooks routing."""


# pylint: disable=unused-import
from .default_router import (  # noqa: F401
    dispatch_event,
    process_event,
    process_event_actions,
    WEBHOOK_EVENTS_ROUTER,
)



"""Test app server machinery."""

import uuid

from aiohttp.client import ClientSession
from aiohttp.web import SockSite
from aiohttp.test_utils import get_unused_port_socket
import pytest

from octomachinery.app.config import BotAppConfig
from octomachinery.app.routing import WEBHOOK_EVENTS_ROUTER
from octomachinery.app.server.machinery import setup_server_runner
from octomachinery.github.api.app_client import GitHubApp


IPV4_LOCALHOST = '127.0.0.1'


@pytest.fixture
def ephemeral_port_tcp_sock():
    """Initialize an ephemeral TCP socket."""
    return get_unused_port_socket(IPV4_LOCALHOST)


@pytest.fixture
def ephemeral_port_tcp_sock_addr(ephemeral_port_tcp_sock):
    """Return final host and port addr of the ephemeral TCP socket."""
    return ephemeral_port_tcp_sock.getsockname()[:2]


@pytest.fixture
def github_app_id() -> int:
    """Return a GitHub App ID."""
    return 0


@pytest.fixture
def octomachinery_config(
        github_app_id: int, rsa_private_key_bytes: bytes,
        ephemeral_port_tcp_sock_addr: tuple,
) -> None:
    """Initialize a GitHub App bot config."""
    host, port = ephemeral_port_tcp_sock_addr
    # https://github.com/hynek/environ-config/blob/master/CHANGELOG.rst#1910-2019-09-02
    return BotAppConfig.from_environ({  # pylint: disable=no-member
        'GITHUB_APP_IDENTIFIER': str(github_app_id),
        'GITHUB_PRIVATE_KEY': rsa_private_key_bytes.decode(),
        'HOST': host,
        'PORT': port,
    })


@pytest.fixture
def octomachinery_config_github_app(octomachinery_config):
    """Return a GitHub App bot config section."""
    return octomachinery_config.github


@pytest.fixture
def octomachinery_config_server(octomachinery_config):
    """Return a GitHub App server config section."""
    return octomachinery_config.server


@pytest.fixture
async def aiohttp_client_session(loop) -> ClientSession:
    """Initialize an aiohttp HTTP client session."""
    async with ClientSession() as http_session:
        yield http_session


@pytest.fixture
def octomachinery_event_routers():
    """Construct a set of routers for use in the GitHub App."""
    return frozenset({WEBHOOK_EVENTS_ROUTER})


@pytest.fixture
def github_app(
        octomachinery_config_github_app, aiohttp_client_session,
        octomachinery_event_routers,
):
    """Initizalize a GitHub App instance."""
    return GitHubApp(
        octomachinery_config_github_app,
        aiohttp_client_session,
        octomachinery_event_routers,
    )


@pytest.fixture
async def octomachinery_app_server_runner(github_app):
    """Set up an HTTP handler for webhooks."""
    return await setup_server_runner(github_app)


@pytest.fixture
async def octomachinery_app_tcp(
        ephemeral_port_tcp_sock,
        octomachinery_app_server_runner,
):
    """Run octomachinery web server and tear-down after testing."""
    tcp_site = SockSite(
        octomachinery_app_server_runner, ephemeral_port_tcp_sock,
    )
    await tcp_site.start()
    try:
        yield tcp_site
    finally:
        await tcp_site.stop()


@pytest.fixture
async def send_webhook_event(
        octomachinery_app_tcp, aiohttp_client_session,
):
    """Return a webhook sender coroutine."""
    def _send_event(webhook_payload=None):
        post_body = {} if webhook_payload is None else webhook_payload

        webhook_endpoint_url = octomachinery_app_tcp.name
        return aiohttp_client_session.post(
            webhook_endpoint_url, json=post_body,
            headers={
                'X-GitHub-Delivery': str(uuid.uuid4()),
                'X-GitHub-Event': 'ping',
            },
        )
    return _send_event


async def test_ping_response(send_webhook_event, github_app_id):
    """Test that ping webhook event requests receive a HTTP response."""
    async with send_webhook_event(
            {
                'hook': {'app_id': github_app_id},
                'hook_id': 0,
                'zen': 'Hey zen!',
            },
    ) as gh_app_http_resp:
        resp_body: str = (await gh_app_http_resp.read()).decode()

    resp_content_type = gh_app_http_resp.headers['Content-Type']
    expected_response_start = (
        'OK: GitHub event received and scheduled for processing. '
        "It is GidgetHubWebhookEvent(name='ping', payload={'hook': {"
        f"'app_id': {github_app_id}"
        "}, 'hook_id': 0, 'zen': 'Hey zen!'}, delivery_id=UUID('"
    )

    assert resp_content_type == 'text/plain; charset=utf-8'
    assert resp_body.startswith(expected_response_start)



"""Web-server constructors."""

import functools
import logging

import anyio
from aiohttp.client import ClientSession
from aiohttp import web

# pylint: disable=relative-beyond-top-level
from ...github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ...utils.asynctools import auto_cleanup_aio_tasks
# pylint: disable=relative-beyond-top-level
from ..routing.webhooks_dispatcher import route_github_webhook_event


logger = logging.getLogger(__name__)


async def start_tcp_site(
        server_config, aiohttp_server_runner: web.ServerRunner,
) -> web.TCPSite:
    """Return initialized and listening TCP site."""
    host, port = server_config.host, server_config.port
    aiohttp_tcp_site = web.TCPSite(aiohttp_server_runner, host, port)
    await aiohttp_tcp_site.start()
    logger.info(
        ' Serving on %s '.center(50, '='),
        aiohttp_tcp_site.name,
    )
    return aiohttp_tcp_site


async def get_server_runner(http_handler):
    """Initialize server runner."""
    aiohttp_server = web.Server(http_handler)
    aiohttp_server_runner = web.ServerRunner(
        aiohttp_server,
        # handle SIGTERM and SIGINT
        # by raising aiohttp.web_runner.GracefulExit exception
        handle_signals=True,
    )
    await aiohttp_server_runner.setup()
    return aiohttp_server_runner


async def _prepare_github_app(github_app):
    """Set GitHub App in the context."""
    logger.info('Starting the following GitHub App:')
    logger.info(
        '* app id: %s',
        github_app._config.app_id,  # pylint: disable=protected-access
    )
    logger.info(
        '* private key SHA-1 fingerprint: %s',
        # pylint: disable=protected-access
        github_app._config.private_key.fingerprint,
    )
    logger.info(
        '* user agent: %s',
        github_app._config.user_agent,  # pylint: disable=protected-access
    )
    await github_app.log_installs_list()


async def _launch_web_server_and_wait_until_it_stops(
        web_server_config,
        github_app: GitHubApp,
        webhook_secret: str = None,
) -> None:
    """Start a web server.

    And then block until SIGINT comes in.
    """
    aiohttp_server_runner = await setup_server_runner(
        github_app, webhook_secret,
    )
    aiohttp_tcp_site = await start_tcp_site(
        web_server_config, aiohttp_server_runner,
    )
    await _stop_site_on_cancel(aiohttp_tcp_site)


async def setup_server_runner(
        github_app: GitHubApp,
        webhook_secret: str = None,
) -> web.ServerRunner:
    """Return a server runner with a webhook dispatcher set up."""
    return await get_server_runner(
        functools.partial(
            route_github_webhook_event,
            github_app=github_app,
            webhook_secret=webhook_secret,
        ),
    )


async def _stop_site_on_cancel(aiohttp_tcp_site):
    """Stop the server after SIGINT."""
    try:
        await anyio.sleep(float('inf'))
    except anyio.get_cancelled_exc_class():
        logger.info(' Stopping the server '.center(50, '='))
        await aiohttp_tcp_site.stop()


def log_webhook_secret_status(webhook_secret):
    """Log HTTP body signature verification behavior."""
    webhook_secret_repr = (
        f' ({webhook_secret[:1]}...{webhook_secret[-1:]})'
        if webhook_secret else ''
    )
    logger.info(
        'Webhook secret%s is [%sSET]: %s',
        webhook_secret_repr,
        '' if webhook_secret else 'NOT ',
        'SIGNATURE VERIFICATION WILL BE ENFORCED'
        if webhook_secret else 'SIGNED WEBHOOKS WILL BE REJECTED',
    )


@auto_cleanup_aio_tasks
async def run_forever(config, event_routers):
    """Spawn an HTTP server in anyio context."""
    logger.debug('The GitHub App env is set to `%s`', config.runtime.env)
    log_webhook_secret_status(config.github.webhook_secret)
    async with ClientSession() as aiohttp_client_session:
        github_app = GitHubApp(
            config.github,
            http_session=aiohttp_client_session,
            event_routers=event_routers,
        )
        await _prepare_github_app(github_app)
        await _launch_web_server_and_wait_until_it_stops(
            config.server, github_app, config.github.webhook_secret,
        )



#! /usr/bin/env python3
"""Octomachinery CLI entrypoint."""

import asyncio
from functools import wraps
import importlib
from io import TextIOWrapper
import os
import pathlib
from typing import Any, Callable, Iterable, Iterator, Set

from aiohttp.client import ClientSession
import click

# pylint: disable=relative-beyond-top-level
from ..app.config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from ..app.routing.abc import OctomachineryRouterBase
# pylint: disable=relative-beyond-top-level
from ..app.routing.webhooks_dispatcher import route_github_event
# pylint: disable=relative-beyond-top-level
from ..github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ..github.entities.action import GitHubAction
# pylint: disable=relative-beyond-top-level
from ..github.models.events import GitHubEvent
# pylint: disable=relative-beyond-top-level
from ..github.models.events import GitHubWebhookEvent


@click.group()
@click.pass_context
def cli(ctx: click.Context) -> None:  # pylint: disable=unused-argument
    """Click CLI base."""


def run_async(orig_async_func: Callable[..., Any]):
    """Run the given async func in event loop."""
    @wraps(orig_async_func)
    def func_wrapper(*args: Any, **kwargs: Any):
        return asyncio.run(orig_async_func(*args, **kwargs))
    return func_wrapper


@cli.command()
@click.option('--event', '-e', prompt=False, type=str)
@click.option(
    '--payload-path', '-p',
    'event_payload',
    prompt=True,
    type=click.File(mode='r'),
)
@click.option('--token', '-t', prompt=False, type=str)
@click.option('--app', '-a', prompt=False, type=int)
@click.option('--private-key', '-P', prompt=False, type=click.File(mode='r'))
@click.option('--entrypoint-module', '-m', prompt=False, type=str)
@click.option(
    '--event-router', '-r',
    'event_routers',
    multiple=True,
    prompt=False,
    type=str,
)
@click.pass_context
@run_async
async def receive(  # pylint: disable=too-many-arguments,too-many-locals
        ctx: click.Context,
        event: str, event_payload: TextIOWrapper,
        token: str,
        app: int, private_key: TextIOWrapper,
        entrypoint_module: str,
        event_routers: Iterable[str],
) -> None:
    """Webhook event receive command."""
    app_missing_private_key = app is not None and not private_key
    if app_missing_private_key:
        ctx.fail(click.style('App requires a private key', fg='red'))

    creds_present = token or (app and private_key)
    if not creds_present:
        ctx.fail(click.style('GitHub auth credentials are missing', fg='red'))

    too_many_creds_present = token and (app or private_key)
    if too_many_creds_present:
        ctx.fail(
            click.style(
                'Please choose between a token or '
                'an app id with a private key',
                fg='red',
            ),
        )

    make_event = GitHubEvent if app is None else GitHubWebhookEvent
    try:
        gh_event = make_event.from_fixture_fd(event_payload, event=event)
    except ValueError as val_err:
        ctx.fail(click.style(str(val_err), fg='red'))

    os.environ.update(get_extra_env_vars(gh_event, token, app, private_key))

    try:
        target_routers = set(
            load_event_routers(entrypoint_module, event_routers),
        )
    except AttributeError as attr_err:
        ctx.fail(
            click.style(
                f'Could not find an event router: {attr_err!s}',
                fg='red',
            ),
        )
    except ImportError as imp_err:
        ctx.fail(
            click.style(f'Could not load a module: {imp_err!s}', fg='red'),
        )

    config = BotAppConfig.from_dotenv()
    gh_app_kwargs = {'config': config.github}
    make_gh_app = GitHubApp
    if app is None:
        make_gh_app = GitHubAction
        gh_app_kwargs['metadata'] = config.action
    async with ClientSession() as http_client_session:
        github_app = make_gh_app(
            http_session=http_client_session,
            event_routers=target_routers or None,
            **gh_app_kwargs,
        )
        await route_github_event(
            github_event=gh_event, github_app=github_app,
        )

    click.echo(
        click.style(
            f'Finished processing {gh_event.name!s} event!',
            fg='green',
        ),
    )


def load_event_routers(
        entrypoint_module: str = None,
        event_routers: Set[str] = frozenset(),
) -> Iterator[OctomachineryRouterBase]:
    """Yield event routers from strings."""
    if entrypoint_module is not None:
        importlib.import_module(entrypoint_module)

    for router_path in event_routers:
        target_sep = ':' if ':' in router_path else '.'
        module_path, _sep, target_router = router_path.rpartition(target_sep)
        yield getattr(importlib.import_module(module_path), target_router)


def get_extra_env_vars(
        gh_event: GitHubEvent, token: str, app: int,
        private_key: TextIOWrapper,
) -> dict:
    """Construct additional env vars for App or Action processing."""
    env = {}

    if app is not None:
        env['OCTOMACHINERY_APP_MODE'] = 'app'

        env['GITHUB_APP_IDENTIFIER'] = str(app)
        env['GITHUB_PRIVATE_KEY'] = private_key.read()
        return env

    env['OCTOMACHINERY_APP_MODE'] = 'action'

    env['GITHUB_ACTION'] = 'Fake CLI Action'
    env['GITHUB_ACTOR'] = gh_event.payload.get('sender', {}).get('login', '')
    env['GITHUB_EVENT_NAME'] = gh_event.name
    env['GITHUB_WORKSPACE'] = str(pathlib.Path('.').resolve())
    env['GITHUB_SHA'] = gh_event.payload.get('head_commit', {}).get('id', '')
    env['GITHUB_REF'] = gh_event.payload.get('ref', '')
    env['GITHUB_REPOSITORY'] = (
        gh_event.payload.
        get('repository', {}).
        get('full_name', '')
    )
    env['GITHUB_TOKEN'] = token
    env['GITHUB_WORKFLOW'] = 'Fake CLI Workflow'
    env['GITHUB_EVENT_PATH'] = '/dev/null'

    return env


def main():
    """CLI entrypoint."""
    kwargs = {
        'prog_name': f'python3 -m {__package__}',
    } if __name__ == '__main__' else {}

    return cli(  # pylint: disable=unexpected-keyword-arg
        auto_envvar_prefix='OCTOMACHINERY_CLI_',
        obj={},
        **kwargs,
    )


__name__ == '__main__' and main()  # pylint: disable=expression-not-assigned



"""Tests for circular imports in all local packages and modules.

This ensures all internal packages can be imported right away without
any need to import some other module before doing so.

This module is based on an idea that pytest uses for self-testing:
* https://github.com/pytest-dev/pytest/blob/d18c75b/testing/test_meta.py
* https://twitter.com/codewithanthony/status/1229445110510735361
"""
from itertools import chain
from pathlib import Path
import os
import pkgutil
import subprocess
import sys

import pytest

import octomachinery


def _find_all_importables(pkg):
    """Find all importables in the project.

    Return them in order.
    """
    return sorted(
        set(
            chain.from_iterable(
                _discover_path_importables(Path(p), pkg.__name__)
                for p in pkg.__path__
            ),
        ),
    )


def _discover_path_importables(pkg_pth, pkg_name):
    """Yield all importables under a given path and package."""
    for dir_path, _d, file_names in os.walk(pkg_pth):
        pkg_dir_path = Path(dir_path)

        if pkg_dir_path.parts[-1] == '__pycache__':
            continue

        if all(Path(_).suffix != '.py' for _ in file_names):
            continue

        rel_pt = pkg_dir_path.relative_to(pkg_pth)
        pkg_pref = '.'.join((pkg_name, ) + rel_pt.parts)
        yield from (
            pkg_path
            for _, pkg_path, _ in pkgutil.walk_packages(
                (str(pkg_dir_path), ), prefix=f'{pkg_pref}.',
            )
        )


@pytest.mark.parametrize(
    'import_path',
    _find_all_importables(octomachinery),
)
def test_no_warnings(import_path):
    """Verify that exploding importables doesn't explode.

    This is seeking for any import errors including ones caused
    by circular imports.
    """
    imp_cmd = (
        sys.executable,
        '-W', 'error',
        '-W', 'ignore:"@coroutine" decorator is deprecated '
        'since Python 3.8, use "async def" instead:'
        'DeprecationWarning:aiohttp.helpers',
        '-c', f'import {import_path!s}',
    )

    subprocess.check_call(imp_cmd)



"""Asynchronous tools set."""

from functools import wraps
from operator import itemgetter

from anyio import create_queue, create_task_group as all_subtasks_awaited


def auto_cleanup_aio_tasks(async_func):
    """Ensure all subtasks finish."""
    @wraps(async_func)
    async def async_func_wrapper(*args, **kwargs):
        async with all_subtasks_awaited():
            return await async_func(*args, **kwargs)
    return async_func_wrapper


async def _send_task_res_to_q(res_q, task_id, aio_task):
    """Await task and put its result to the queue."""
    try:
        task_res = await aio_task
    except (BaseException, Exception) as exc:
        task_res = exc
        raise
    finally:
        await res_q.put((task_id, task_res))


async def _aio_gather_iter_pairs(*aio_tasks):
    """Spawn async tasks and yield with pairs of ids with results."""
    aio_tasks_num = len(aio_tasks)
    task_res_q = create_queue(aio_tasks_num)

    async with all_subtasks_awaited() as task_group:
        for task_id, task in enumerate(aio_tasks):
            await task_group.spawn(
                _send_task_res_to_q,
                task_res_q,
                task_id, task,
            )

        for _ in range(aio_tasks_num):
            yield await task_res_q.get()


async def aio_gather_iter(*aio_tasks):
    """Spawn async tasks and yield results."""
    async for _task_id, task_res in _aio_gather_iter_pairs(*aio_tasks):
        yield task_res


async def aio_gather(*aio_tasks):
    """Spawn async tasks and return results in the same order."""
    result_pairs_gen = [_r async for _r in _aio_gather_iter_pairs(*aio_tasks)]
    sorted_result_pairs = sorted(result_pairs_gen, key=itemgetter(0))
    all_task_results = map(itemgetter(1), sorted_result_pairs)
    return tuple(all_task_results)


async def try_await(potentially_awaitable):
    """Try awaiting the arg and return it regardless."""
    valid_exc_str = (
        "can't be used in 'await' expression"
    )

    try:
        return await potentially_awaitable
    except TypeError as type_err:
        type_err_msg = str(type_err)
        if not (
                type_err_msg.startswith('object ')
                and type_err_msg.endswith(valid_exc_str)
        ):
            raise

    return potentially_awaitable


async def amap(callback, async_iterable):
    """Map asyncronous generator with a coroutine or a function."""
    async for async_value in async_iterable:
        yield await try_await(callback(async_value))


def dict_to_kwargs_cb(callback):
    """Return a callback mapping dict to keyword arguments."""
    async def callback_wrapper(args_dict):
        return await try_await(callback(**args_dict))
    return callback_wrapper



"""Octomachinery event dispatchers collection."""

import asyncio
from contextlib import suppress
from typing import Any, Iterator, Set, Union

from gidgethub.routing import AsyncCallback, Router as _GidgetHubRouter

from ..utils.asynctools import aio_gather
from ..github.models.events import (
    GidgetHubWebhookEvent, GitHubEvent,
    _GidgetHubEvent,
)
from .abc import OctomachineryRouterBase


__all__ = (
    'GidgetHubRouterBase',
    'ConcurrentRouter',
    'NonBlockingConcurrentRouter',
)


class GidgetHubRouterBase(_GidgetHubRouter, OctomachineryRouterBase):
    """GidgetHub-based router exposing callback matching separately."""

    def emit_routes_for(
            self, event_name: str, event_payload: Any,
    ) -> Iterator[AsyncCallback]:
        """Emit callbacks that match given event and payload.

        :param str event_name: name of the GitHub event
        :param str event_payload: details of the GitHub event

        :yields: coroutine event handlers
        """
        with suppress(KeyError):
            yield from self._shallow_routes[event_name]

        try:
            deep_routes = self._deep_routes[event_name]
        except KeyError:
            return

        for payload_key, payload_values in deep_routes.items():
            if payload_key not in event_payload:
                continue
            event_value = event_payload[payload_key]
            if event_value not in payload_values:
                continue
            yield from payload_values[event_value]

    async def dispatch(
            self, event: Union[GidgetHubWebhookEvent, _GidgetHubEvent],
            *args: Any, **kwargs: Any,
    ) -> None:
        """Invoke handler tasks for the given event sequentially."""
        if isinstance(event, _GidgetHubEvent):
            event = GidgetHubWebhookEvent.from_gidgethub(event)

        callback_gen = self.emit_routes_for(event.name, event.payload)
        callback_coros = (cb(event, *args, **kwargs) for cb in callback_gen)

        for coro in callback_coros:
            await coro


class ConcurrentRouter(GidgetHubRouterBase):
    """GitHub event router invoking event handlers simultaneously."""

    async def dispatch(
            self, event: GitHubEvent,
            *args: Any, **kwargs: Any,
    ) -> None:
        """Invoke coroutine callbacks for the given event together."""
        callback_gen = self.emit_routes_for(event.name, event.payload)
        callback_coros = (cb(event, *args, **kwargs) for cb in callback_gen)

        await aio_gather(*callback_coros)


class NonBlockingConcurrentRouter(ConcurrentRouter):
    """Non-blocking GitHub event router scheduling handler tasks."""

    def __init__(self, *args, **kwargs):
        """Initialize NonBlockingConcurrentRouter."""
        super().__init__(*args, **kwargs)
        # NOTE: For some reason, mypy doesn't accept anything except Any here:
        self._event_handler_tasks: Set[Any] = set()

    async def dispatch(
            self, event: GitHubEvent,
            *args: Any, **kwargs: Any,
    ) -> None:
        """Schedule coroutine callbacks for the given event together."""
        callback_gen = self.emit_routes_for(event.name, event.payload)
        callback_coros = (cb(event, *args, **kwargs) for cb in callback_gen)
        handler_tasks = map(asyncio.create_task, callback_coros)
        self._event_handler_tasks.update(handler_tasks)
